.imp-status-menu-button {
  padding: 1px 8px 2px 4px;
  border: 0;
  outline-width: 0;
}

.imp-statusbar {
  border-top: 1px solid @borders;
  padding-left: 10px;
}

.imp-settings-overlay {
  border-radius: 0px;
}

.imp-tool-bar {
  border-radius: 0px;
}

.imp-rule-editor-tiny-button-row {
  min-width: 0px;
  padding: 0px;
}

.imp-rule-editor-tiny-button-column {
  min-height: 0px;
  padding-bottom: 0px;
  padding-top: 0px;
}

.bus-toolbar {
  border-right: 1px solid @borders;
}

.tag-entry-tiny-button {
  min-width: 0px;
  padding: 0px;
}

.imp-search-slider {
  background-color: @theme_base_color;
  border-color: @borders;
  border-radius: 0 0 3px 3px;
  border-width: 0 1px 1px 1px;
  border-style: solid;
}

.imp-selection-filter-tiny-button {
  min-width: 0px;
  min-height: 0px;
  padding: 0px;
}

.app-notification *:link {
  color:#4CC6EF;
}

.horizon-scale-label {
  border-radius:0px;
  border-top-left-radius: 5px;
}

.editor-header > *:not(:first-child) {
  border-left: 1px solid rgba(0,0,0,.1);
}

.editor-header label {
  padding: 5px 5px 5px 10px;
}

.editor-header {
  border-top: 1px solid rgba(0,0,0,.1);
  border-bottom: 1px solid rgba(0,0,0,.05);
}

.help-button {
  min-height: 0px;
  min-width: 0px;
  padding: 2px;
}

.imp-tool-popover {
  padding-bottom: 0px;
}

.imp-action-bar {
  border-radius:0px;
  border-top-right-radius: 5px;
  border-bottom-right-radius: 5px;
}

.imp-action-bar-expander {
  padding:0px;
  min-width: 0px;
  min-height: 0px;
}

.imp-key-box {
  font-weight: bold;
  border: 2px solid;
  border-radius: 5px;
}

.imp-key-box > label {
 padding-left:.25em;
 padding-right:.25em;
 min-width:1em;
}

.imp-key-box > label:not(:last-child) {
  border-right: 2px solid;
}

.rule-import-header > *:not(:first-child) {
  border-left: 1px solid rgba(0,0,0,.1);
}

.rule-import-header label {
  padding: 5px 10px 5px 10px;
}

.rule-import-header {
  border-bottom: 1px solid rgba(0,0,0,.05);
}

.imp-symbol-expanded-preview {
  border-color: @borders;
  border-radius: 0 0 3px 0px;
  border-width: 0 1px 1px 0px;
  border-style: solid;
  padding: 10px;
}
